// The MIT License

// Copyright (c) 2009 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

import Connectable::*;
import GetPut::*;
import ClientServer::*;
import RegFile::*;
import FIFO::*;
import FIFOF::*;
import RWire::*;

import BFIFO::*;
import MemTypes::*;
import ProcTypes::*;
import Trace::*;

interface ProcStats;
  interface Get#(Stat) num_cycles;
  interface Get#(Stat) num_inst;
endinterface

interface CPUToHost;
  method Bit#(32) cpuToHost(int req);
endinterface

interface Proc;

  // Interface from processor to caches
  interface Client#(DataReq,DataResp) dmem_client;
  interface Client#(InstReq,InstResp) imem_client;

  // Interface for enabling/disabling statistics on the rest of the core
  interface Get#(Bool) statsEn_get;

  // Interface for collecting statistics.
  interface ProcStats stats;

  // Interface to host
  interface CPUToHost tohost;

endinterface


typedef enum { PCgen, Exec, Writeback } Stage deriving(Eq,Bits);

//-----------------------------------------------------------
// Register file module
//-----------------------------------------------------------

interface RFile;
   method Action   wr( Rindx rindx, Bit#(32) data );
   method Bit#(32) rd1( Rindx rindx );
   method Bit#(32) rd2( Rindx rindx );
endinterface

module mkRFile( RFile );
   
   RegFile#(Rindx,Bit#(32)) rfile <- mkRegFileFull();
   
   method Action wr( Rindx rindx, Bit#(32) data );
      rfile.upd( rindx, data );
   endmethod
   
   method Bit#(32) rd1( Rindx rindx );
      return ( rindx == 0 ) ? 0 : rfile.sub(rindx);
   endmethod
   
   method Bit#(32) rd2( Rindx rindx );
      return ( rindx == 0 ) ? 0 : rfile.sub(rindx);
   endmethod

endmodule

//-----------------------------------------------------------
// Helper functions
//-----------------------------------------------------------

function Bit#(32) slt( Bit#(32) val1, Bit#(32) val2 );
   return zeroExtend( pack( signedLT(val1,val2) ) );
endfunction

function Bit#(32) sltu( Bit#(32) val1, Bit#(32) val2 );
   return zeroExtend( pack( val1 < val2 ) );
endfunction

function Bit#(32) rshft( Bit#(32) val );
   return zeroExtend(val[4:0]);
endfunction

//-----------------------------------------------------------
// Reference processor
//-----------------------------------------------------------


(* doc = "synthesis attribute ram_style mkProc distributed;" *)
(* synthesize *)
module  mkProc( Proc );

   //-----------------------------------------------------------
   // State

   // Standard processor state
   
   Reg#(Addr)  pc    <- mkReg(32'h00001000);
   Reg#(Stage) stage <- mkReg(PCgen);
   RFile       rf    <- mkRFile;

   Reg#(Bit#(32)) cp0_tohost   <- mkReg(0);
   Reg#(Bit#(32)) cp0_fromhost <- mkReg(0);
   Reg#(Bool)    cp0_statsEn  <- mkReg(False);

   // Memory request/response state
   
   FIFO#(InstReq)  instReqQ    <- mkBFIFO1();
   FIFO#(InstResp) instRespQ   <- mkFIFO();

   FIFO#(DataReq)  dataReqQ    <- mkBFIFO1();
   FIFO#(DataResp) dataRespQ   <- mkFIFO();

   // Statistics state
   Reg#(Stat) num_cycles <- mkReg(0);
   Reg#(Stat) num_inst <- mkReg(0);

   //-----------------------------------------------------------
   // Rules

   let pc_plus4 = pc + 4;

   rule pcgen ( stage == PCgen );
      traceTiny("mkProc", "pc",pc);
      traceTiny("mkProc", "pcgen","P");
      instReqQ.enq( LoadReq{ addr:pc, tag:0 } );
      stage   <= Exec;
   endrule

   rule exec ( stage == Exec );

      // Some abbreviations
      let sext = signExtend;
      let zext = zeroExtend;
      let sra  = signedShiftRight;
      
      // Get the instruction
      
      instRespQ.deq();
      Instr inst 
      = case ( instRespQ.first() ) matches
	   tagged LoadResp  .ld : return unpack(ld.data);
	   tagged StoreResp .st : return ?;
	endcase;

      // Some default variables
      Stage next_stage = PCgen;
      Addr  next_pc    = pc_plus4;

      // Tracing
      traceTiny("mkProc", "exec","X");
      traceTiny("mkProc", "exInstTiny",inst);
      traceFull("mkProc", "exInstFull",inst);

      case ( inst ) matches

	 // -- Memory Ops ------------------------------------------------      

	 tagged LW .it : 
	    begin
               Addr addr = rf.rd1(it.rbase) + sext(it.offset);
               dataReqQ.enq( LoadReq{ addr:addr, tag:zeroExtend(it.rdst) } );
               next_stage = Writeback;
	    end

	 tagged SW .it : 
	    begin
               Addr addr = rf.rd1(it.rbase) + sext(it.offset);
               dataReqQ.enq( StoreReq{ tag:0, addr:addr, data:rf.rd2(it.rsrc) } );
               next_stage = Writeback;
	    end

	 // -- Simple Ops ------------------------------------------------      

	 tagged ADDIU .it : rf.wr( it.rdst, rf.rd1(it.rsrc) + sext(it.imm) );
	 tagged SLTI  .it : rf.wr( it.rdst, slt( rf.rd1(it.rsrc), sext(it.imm) ) );
	 tagged SLTIU .it : rf.wr( it.rdst, sltu( rf.rd1(it.rsrc), sext(it.imm) ) );
	 tagged ANDI  .it : 
	    begin
	       Bit#(32) zext_it_imm = zext(it.imm);
	       rf.wr( it.rdst, rf.rd1(it.rsrc) & zext_it_imm );
	    end
	 tagged ORI   .it : 
	    begin
	       Bit#(32) zext_it_imm = zext(it.imm);
	       rf.wr( it.rdst, rf.rd1(it.rsrc) | zext_it_imm );
	    end
	 tagged XORI  .it : 
	    begin
	       Bit#(32) zext_it_imm = zext(it.imm);
	       rf.wr( it.rdst, rf.rd1(it.rsrc) ^ zext_it_imm );
	    end
	 tagged LUI   .it : 
	    begin
	       Bit#(32) zext_it_imm = zext(it.imm);
	       rf.wr( it.rdst, (zext_it_imm << 32'd16) );
	    end
	 
	 tagged SLL   .it : 
	    begin
	       Bit#(32) zext_it_shamt = zext(it.shamt);
	       rf.wr( it.rdst, rf.rd1(it.rsrc) << zext_it_shamt );
	    end
	 tagged SRL   .it : 
	    begin
	       Bit#(32) zext_it_shamt = zext(it.shamt);
	       rf.wr( it.rdst, rf.rd1(it.rsrc) >> zext_it_shamt );
	    end
	 tagged SRA   .it : 
	    begin
	       Bit#(32) zext_it_shamt = zext(it.shamt);
	       rf.wr( it.rdst, sra( rf.rd1(it.rsrc), zext_it_shamt ));
	    end
	 tagged SLLV  .it : rf.wr( it.rdst, rf.rd1(it.rsrc) << rshft(rf.rd2(it.rshamt)) );
	 tagged SRLV  .it : rf.wr( it.rdst, rf.rd1(it.rsrc) >> rshft(rf.rd2(it.rshamt)) );
	 tagged SRAV  .it : rf.wr( it.rdst, sra( rf.rd1(it.rsrc), rshft(rf.rd2(it.rshamt)) ) );
	 tagged ADDU  .it : rf.wr( it.rdst, rf.rd1(it.rsrc1) + rf.rd2(it.rsrc2) );
	 tagged SUBU  .it : rf.wr( it.rdst, rf.rd1(it.rsrc1) - rf.rd2(it.rsrc2) );
	 tagged AND   .it : rf.wr( it.rdst, rf.rd1(it.rsrc1) & rf.rd2(it.rsrc2) );
	 tagged OR    .it : rf.wr( it.rdst, rf.rd1(it.rsrc1) | rf.rd2(it.rsrc2) );
	 tagged XOR   .it : rf.wr( it.rdst, rf.rd1(it.rsrc1) ^ rf.rd2(it.rsrc2) );
	 tagged NOR   .it : rf.wr( it.rdst, ~(rf.rd1(it.rsrc1) | rf.rd2(it.rsrc2)) );
	 tagged SLT   .it : rf.wr( it.rdst, slt( rf.rd1(it.rsrc1), rf.rd2(it.rsrc2) ) );
	 tagged SLTU  .it : rf.wr( it.rdst, sltu( rf.rd1(it.rsrc1), rf.rd2(it.rsrc2) ) );

	 // -- Branches --------------------------------------------------

	 tagged BLEZ  .it : 
            if ( signedLE( rf.rd1(it.rsrc), 0 ) )
               next_pc = pc_plus4 + (sext(it.offset) << 2);

	 tagged BGTZ  .it : 
            if ( signedGT( rf.rd1(it.rsrc), 0 ) ) 
               next_pc = pc_plus4 + (sext(it.offset) << 2);

	 tagged BLTZ  .it : 
            if ( signedLT( rf.rd1(it.rsrc), 0 ) )
               next_pc = pc_plus4 + (sext(it.offset) << 2);

	 tagged BGEZ  .it : 
            if ( signedGE( rf.rd1(it.rsrc), 0 ) )
               next_pc = pc_plus4 + (sext(it.offset) << 2);

      tagged BEQ   .it : 
        if ( rf.rd1(it.rsrc1) == rf.rd2(it.rsrc2) )
          next_pc = pc_plus4 + (sext(it.offset) << 2);

      tagged BNE   .it : 
        if ( rf.rd1(it.rsrc1) != rf.rd2(it.rsrc2) )
          next_pc = pc_plus4 + (sext(it.offset) << 2);
      
      // -- Jumps -----------------------------------------------------
      
      tagged J     .it : 
        next_pc = { pc_plus4[31:28], it.target, 2'b0 };
      
      tagged JR    .it : 
        next_pc = rf.rd1(it.rsrc);

      tagged JAL   .it : 
       begin
        rf.wr( 31, pc_plus4 );
        next_pc = { pc_plus4[31:28], it.target, 2'b0 };
       end

      tagged JALR  .it : 
       begin
        rf.wr( it.rdst, pc_plus4 );
        next_pc = rf.rd1(it.rsrc);
       end

      // -- Cop0 ------------------------------------------------------
      
      tagged MTC0  .it : 
        case ( it.cop0dst )
          5'd10 : cp0_statsEn <= unpack(truncate(rf.rd1(it.rsrc))); 
          5'd21 : cp0_tohost  <= truncate(rf.rd1(it.rsrc));
          default : 
            $display( " RTL-ERROR : %m : Illegal MTC0 cop0dst register!" );
        endcase

      tagged MFC0  .it : 
        case ( it.cop0src )
          5'd10 : rf.wr( it.rdst, zext(pack(cp0_statsEn)) );
          5'd20 : rf.wr( it.rdst, cp0_fromhost );
          5'd21 : rf.wr( it.rdst, cp0_tohost   );
          default : 
            $display( " RTL-ERROR : %m : Illegal MFC0 cop0src register!" );
        endcase

      // -- Illegal ---------------------------------------------------

      default : 
        $display( " RTL-ERROR : %m : Illegal instruction !" );

    endcase

    stage    <= next_stage;
    pc       <= next_pc;

    if ( cp0_statsEn )
      num_inst <= num_inst+1;

  endrule

  rule writeback ( stage == Writeback );
    traceTiny("mkProc", "writeback","W");

    dataRespQ.deq();
    case ( dataRespQ.first() ) matches
      tagged LoadResp .ld  : rf.wr( truncate(ld.tag), ld.data );	
      tagged StoreResp .st : noAction;
    endcase

    stage <= PCgen;
  endrule

  rule inc_num_cycles;
    if ( cp0_statsEn )
      num_cycles <= num_cycles+1;
  endrule


  //-----------------------------------------------------------
  // Methods

  interface Client imem_client;
    interface Get request  = toGet(instReqQ);
    interface Put response = toPut(instRespQ);
  endinterface

  interface Client dmem_client;
    interface Get request  = toGet(dataReqQ);
    interface Put response = toPut(dataRespQ);
  endinterface

  interface Get statsEn_get = toGet(asReg(cp0_statsEn));

  interface ProcStats stats;
    interface Get num_cycles = toGet(asReg(num_cycles));
    interface Get num_inst = toGet(asReg(num_inst));
  endinterface

  interface CPUToHost tohost;
    method Bit#(32) cpuToHost(int req);
      return (case (req)
       0: cp0_tohost;
       1: pc;
       2: zeroExtend(pack(stage));
      endcase);
    endmethod
  endinterface

endmodule

